/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * This file is part of SableCC.                             *
 * See the file "LICENSE" for copyright information and the  *
 * terms and conditions for copying, distribution and        *
 * modification of SableCC.                                  *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

package org.sablecc.sablecc;

import java.util.*;

final class Symbol implements Comparable<Symbol>
{
  private static List<Symbol> terminals;
  private static List<Symbol> nonterminals;
  private static Map<String, Symbol> names;

  private static boolean modified_ = true;
  private static Symbol[] symbols_;
  private static Symbol[] terminals_;
  private static Symbol[] nonterminals_;

  final String name;
  final String errorName;
  final boolean terminal;
  final int index;

  static {
    reinit();
  }

  Symbol(String name, String errorName, boolean terminal)
  {
    if(names.get(name) != null)
    {
      throw new IllegalArgumentException("The symbol " + name + " aready exists.");
    }

    if(terminal)
    {
      terminals.add(this);
      this.index = terminals.indexOf(this);
    }
    else
    {
      nonterminals.add(this);
      this.index = nonterminals.indexOf(this);
    }

    this.name = name;
    this.errorName = errorName;
    this.terminal = terminal;
    names.put(name, this);
    modified_ = true;
  }

  public static void reinit()
  {
    terminals = new ArrayList<>();
    nonterminals = new ArrayList<>();
    names = new TreeMap<>();
    modified_ = true;
    symbols_ = null;
    terminals_ = null;
    nonterminals_ = null;
  }

  static Symbol symbol(String name)
  {
    return names.get(name);
  }

  static Symbol symbol(int index, boolean terminal)
  {
    if(terminal)
    {
      return terminals.get(index);
    }
    else
    {
      return nonterminals.get(index);
    }
  }

  private static void computeArrays()
  {
    symbols_ = new Symbol[terminals.size() + nonterminals.size()];
    terminals_ = terminals.toArray(new Symbol[0]);
    nonterminals_ = nonterminals.toArray(new Symbol[0]);

    System.arraycopy(terminals_, 0, symbols_, 0, terminals_.length);
    System.arraycopy(nonterminals_, 0, symbols_, terminals_.length, nonterminals_.length);

    modified_ = false;
  }

  static Symbol[] symbols()
  {
    if(modified_)
    {
      computeArrays();
    }

    return symbols_;
  }

  static Symbol[] terminals()
  {
    if(modified_)
    {
      computeArrays();
    }

    return terminals_;
  }

  static Symbol[] nonterminals()
  {
    if(modified_)
    {
      computeArrays();
    }

    return nonterminals_;
  }

  @Override
  public String toString()
  {
    return name;
  }

  @Override
  public int compareTo(Symbol symbol)
  {
    if(terminal ^ symbol.terminal)
    {
      if(terminal)
      {
        return 1;
      }

      return -1;
    }

    return index - symbol.index;
  }
}
